package meerkat.crypto.dkg.feldman;

import com.google.protobuf.InvalidProtocolBufferException;
import meerkat.crypto.utils.Channel;
import meerkat.crypto.secretsharing.shamir.Polynomial;
import com.google.protobuf.ByteString;
import com.google.protobuf.Message;
import meerkat.protobuf.DKG;
import org.factcenter.qilin.primitives.Group;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Set;

import static meerkat.crypto.dkg.comm.MessageUtils.createMessage;

/**
 * Created by Tzlil on 3/14/2016.
 *
 * implementation of joint feldman protocol user.
 *
 * according to the protocol, each user run feldman verifiable secret sharing
 * as a dealer.
 *
 * by the end of run(), each party in QUAL has his own share of the generated random key.
 * this key can be recover by any subset of QUAL of size at least t + 1.
 */
public class User<T> implements Runnable{

    /**
     * joint feldman protocol object
     */
    protected final Protocol<T> dkg;

    /**
     * a generator of cyclic group of order q.
     *          the generated group is a subgroup of the given group.
     *          it must be chosen such that computing discrete logarithms is hard in this group.
     */
    protected final T g;

    /**
     * cyclic group contains g.
     */
    protected final Group<T> group;

    /**
     * user id
     */
    protected final int id;

    /**
     * threshold
     */
    protected final int t;

    /**
     * number of shares
     */
    protected final int n;

    /**
     *  mail handler registered to channel as ReceiverCallback
     */
    protected meerkat.crypto.dkg.comm.MailHandler mailHandler;

    /**
     * channel object
     */
    protected final Channel channel; //

    /**
     * All parties participating in key generation.
     * parties[id-1] has my info.
     */
    protected final Party[] parties;

    /**
     * set of all non-disqualified parties
     */
    protected Set<Integer> QUAL;

    /**
     * my own share of the generated random key.
     */
    protected Polynomial.Point share;

    /**
     * public verification values
     */
    protected ArrayList<T> commitments;

    /**
     * public value,
     * y = g ^ key
     */
    protected T y;

    /**
     * constructor
     * @param dkg joint feldman protocol object
     * @param channel channel object
     */
    public User(Protocol<T> dkg, Channel channel) {
        this.dkg = dkg;

        this.g = dkg.getGenerator();
        this.group = dkg.getGroup();
        this.n = dkg.getN();
        this.t = dkg.getT();
        this.id = dkg.getId();

        this.channel = channel;
        dkg.setChannel(channel);
        registerReceiverCallback();

        this.parties = dkg.getParties();
        this.QUAL = null;
        this.commitments = null;
        this.share = null;
        this.y = null;

    }

    /**
     * create MailHandler and register it as ReceiverCallback
     */
    protected void registerReceiverCallback(){
        this.mailHandler = new MailHandler(new MessageHandler());
        channel.registerReceiverCallback(mailHandler);
    }

    /**
     * stage1 according to the protocol
     *  1. Pi broadcasts Aik for k = 0,...,t.
     *  2. Pi computes the shares Sij for j = 1,...,n and sends Sij secretly to Pj.
     */
    protected void stage1() {
        dkg.broadcastCommitments();
        dkg.sendSecrets();
    }


    /**
     * wait for all shares and commitments will arrive from other parties
     */
    protected void waitUntilStageOneCompleted(){
        // wait for parties' share
        for (int i = 0 ; i < n ; i++){
            synchronized (parties[i]) {
                while (parties[i].share == null && !parties[i].aborted) {
                    try {
                        parties[i].wait();
                    } catch (InterruptedException e) {
                        if (stop) return;
                    }
                }
            }
        }
        // wait for parties' commitments
        for (int i = 0 ; i < n ; i++){
            for (int k = 0 ; k <=  t ; k++) {
                synchronized (parties[i]) {
                    while (parties[i].commitments.get(k) == null && !parties[i].aborted) {
                        try {
                            parties[i].wait();
                        } catch (InterruptedException e) {
                            if (stop) return;
                        }
                    }
                }
            }
        }
    }

    /**
     * stage2 according to the protocol
     *  Pj verifies all the shares he received
     *  if check fails for an index i, Pj broadcasts a complaint against Pi.
     *  Pj broadcasts done message at the end of this stage
     */
    protected void stage2() {
        dkg.broadcastComplaints();
        //broadcast done message after all complaints
        channel.broadcastMessage(createMessage(DKG.Payload.Type.DONE));
    }


    /**
     * wait until all other parties done complaining by receiving done message
     */
    protected void waitUntilStageTwoCompleted(){
        for (int i = 0 ; i < n ; i++){
            synchronized (parties[i]) {
                while (!parties[i].doneFlag && !parties[i].aborted) {
                    try {
                        parties[i].wait();
                    } catch (InterruptedException e) {
                        if (stop) return;
                    }
                }
            }
        }
    }


    /**
     * stage3 according to the protocol
     *  1. if more than t players complain against a player Pi he is disqualified.
     *     otherwise Pi broadcasts the share Sij for each complaining player Pj.
     *  2. if any of the revealed shares fails the verification test, player Pi is disqualified.
     *     set QUAL to be the set of non-disqualified players.
     */
    protected void stage3(){
        dkg.answerAllComplainingPlayers();
        // wait until there is no complaint waiting for answer
        for (int i = 0; i < n; i++){
            for (int j = 0; j < n; j++){
                synchronized (parties[i]) {
                    while (parties[i].complaints[j].equals(Protocol.ComplaintState.Waiting) && !parties[i].aborted) {
                        try {
                            parties[i].wait();
                        } catch (InterruptedException e) {
                            if (stop) return;
                        }
                    }
                }
            }
        }
        this.QUAL = dkg.calcQUAL();
    }

    /**
     * stage4 according to the protocol
     *  1. public value y is computed as y = multiplication of yi mod p for i in QUAL
     *  2. public verification values are computed as Ak = multiplication of Aik mod p for i in QUAL for k = 0,...,t
     *  3. Pj sets is share of the secret as xj = sum of Sij mod q for i in QUAL
     */
    protected void stage4(){
        this.y = dkg.calcY(QUAL);
        this.commitments = dkg.calcCommitments(QUAL);
        this.share = dkg.calcShare(QUAL);
    }

    @Override
    public void run() {
        this.runThread = Thread.currentThread();
        stage1();
        waitUntilStageOneCompleted();
        if (stop) return;
        stage2();
        waitUntilStageTwoCompleted();
        if (stop) return;
        stage3();
        if (stop) return;
        stage4();
    }

    /**
     * current thread in the main loop
     */
    protected Thread runThread;

    /**
     * flag indicates if there was request to stop the current run of the protocol
     */
    protected boolean stop = false;

    /**
     * Request the current run loop to exit gracefully
     */
    public void stop() {
        try {
            stop = true;
            runThread.interrupt();
        }catch (Exception e){
            //do nothing
        }

    }

    /**
     * getter
     * @return commitments
     */
    public ArrayList<T> getCommitments() {
        return commitments;
    }

    /**
     * getter
     * @return g
     */
    public T getGenerator() {
        return g;
    }

    /**
     * getter
     * @return group
     */
    public Group<T> getGroup() {
        return group;
    }

    /**
     * getter
     * @return share
     */
    public Polynomial.Point getShare() {
        return share;
    }

    /**
     * getter
     * @return id
     */
    public int getID() {
        return id;
    }

    /**
     * getter
     * @return n
     */
    public int getN() {
        return n;
    }

    /**
     * getter
     * @return t
     */
    public int getT() {
        return t;
    }

    /**
     * getter
     * @return y
     */
    public T getPublicValue() {
        return y;
    }

    /**
     * getter
     * @return QUAL
     */
    public Set<Integer> getQUAL() {
        return QUAL;
    }

    /**
     * getter
     * @return channel
     */
    public Channel getChannel() {
        return channel;
    }

    /**
     * an implementation of MessageHandler
     */
    public class MessageHandler implements meerkat.crypto.dkg.comm.MessageHandler {

        /**
         * commitment message is valid if:
         *  1. it was received in broadcast chanel
         *  2. the sender didn't sent this commitment before
         */
        protected boolean isValidCommitmentMessage(int sender, boolean isBroadcast, DKG.CommitmentMessage commitmentMessage){
            int i = sender - 1;
            int k = commitmentMessage.getK();
            return isBroadcast && parties[i].commitments.get(k) == null;
        }

        /**
         * secret message is valid if:
         *  1. it was received in private chanel
         *  2. the sender didn't sent secret message before
         *  3. secret.i == i
         *  4. secret.j == id
         */
        protected boolean isValidSecretMessage(int sender, boolean isBroadcast, DKG.ShareMessage secretMessage){
            int i = secretMessage.getI();
            int j = secretMessage.getJ();
            if(sender != i  || isBroadcast)
                return false;
            else
                return parties[i - 1].share == null && j == id;

        }

        /**
         * done message is valid if:
         *  1. it was received in broadcast chanel
         *  2. the sender didn't sent done message before
         */
        protected boolean isValidDoneMessage(int sender, boolean isBroadcast){
            return isBroadcast && !parties[sender - 1].doneFlag;
        }


        /**
         * complaint message is valid if:
         *  1. it was received in broadcast chanel
         *  2. the sender didn't complained against id before
         */
        protected boolean isValidComplaintMessage(int sender, boolean isBroadcast, DKG.IDMessage complaintMessage){
            int i = sender;
            int j = complaintMessage.getId();
            return isBroadcast && parties[i - 1].complaints[j - 1].equals( Protocol.ComplaintState.OK);
        }

        /**
         * answer message is valid if:
         *  1. it was received in broadcast chanel
         *  2. secret.i == i
         *  3. 1 <= secret.j <= n
         *  4. it is marked that j complained against i and i didn't received
         */
        protected boolean isValidAnswerMessage(int sender, boolean isBroadcast, DKG.ShareMessage secretMessage){
            int i = secretMessage.getI();
            int j = secretMessage.getJ();
            if(sender != i  || !isBroadcast)
                return false;
            else
                return j >= 1 && j <= n && parties[i - 1].complaints[j - 1].equals(Protocol.ComplaintState.Waiting);
        }


        @Override
        public void handleMessage(DKG.BroadcastMessage envelope) throws InvalidProtocolBufferException {
            int sender = envelope.getSender();
            boolean isBroadcast = !envelope.getIsPrivate();
            DKG.Payload msg = DKG.Payload.parseFrom(envelope.getPayload());

            switch (msg.getType()) {
                case COMMITMENT:
                    /**
                     * saves the commitment
                     */
                    DKG.CommitmentMessage commitmentMessage = msg.getCommitment();
                    if (isValidCommitmentMessage(sender, isBroadcast, commitmentMessage)) {
                        int i = sender - 1;
                        int k = commitmentMessage.getK();
                        synchronized (parties[i]) {
                            parties[i].commitments.set(k, extractCommitment(commitmentMessage));
                            parties[i].notify();
                        }
                    }
                    break;


                case SHARE:
                    /**
                     * saves the secret
                     */
                    DKG.ShareMessage secretMessage = msg.getShare();
                    if(isValidSecretMessage(sender,isBroadcast,secretMessage)) {
                        int i = secretMessage.getI();
                        Polynomial.Point secret = extractShare(id,secretMessage.getShare());
                        synchronized (parties[i -1]) {
                            parties[i - 1].share = secret;
                            parties[i - 1].notify();
                        }
                    }
                    break;

                case DONE:

                    /**
                     * marks that the sender was finished sending all his complaints
                     */
                    if(isValidDoneMessage(sender,isBroadcast)) {
                        synchronized (parties[sender - 1]) {
                            parties[sender - 1].doneFlag = true;
                            parties[sender - 1].notify();
                        }
                    }
                    break;

                case COMPLAINT:
                    /**
                     * marks that the sender was complained against id
                     */
                    DKG.IDMessage complaintMessage = msg.getId();
                    if(isValidComplaintMessage(sender,isBroadcast,complaintMessage)){
                        int i = sender;
                        int j = complaintMessage.getId();
                        synchronized (parties[j - 1]) {
                            parties[j - 1].complaints[i - 1] = Protocol.ComplaintState.Waiting;
                            parties[j - 1].notify();
                        }
                    }
                    break;
                case ANSWER:
                    /**
                     * if the secret is valid, marks the complaint as NonDisqualified
                     * else marks it as Disqualified
                     * in case that the complainer is id ( j == id ), saves the secret
                     */
                    secretMessage = msg.getShare();
                    if(isValidAnswerMessage(sender,isBroadcast,secretMessage)) {
                        int i = secretMessage.getI();
                        int j = secretMessage.getJ();
                        Polynomial.Point secret = extractShare(j,secretMessage.getShare());
                        synchronized (parties[i - 1]) {
                            if (dkg.isValidShare(secret, parties[i - 1].commitments, j)) {
                                parties[i - 1].complaints[j - 1] = Protocol.ComplaintState.NonDisqualified;
                            } else {
                                parties[i - 1].complaints[j - 1] = Protocol.ComplaintState.Disqualified;
                            }
                            if (j == id) {
                                parties[i - 1].share = secret;
                            }
                            parties[i - 1].notify();
                        }
                    }
                    break;
                case ABORT:
                    /**
                     * marks that the sender was aborted
                     */
                    synchronized (parties[sender - 1]) {
                        parties[sender - 1].aborted = true;
                        parties[sender - 1].notify();
                    }
                    break;



            }
        }

        /**
         * extract share value from ByteString
         * @param i
         * @param share
         * @return new Point (i,share)
         */
        public Polynomial.Point extractShare(int i, ByteString share){
            BigInteger x = BigInteger.valueOf(i);
            BigInteger y = new BigInteger(share.toByteArray());
            return new Polynomial.Point(x,y);
        }

        /**
         *
         * @param commitmentMessage
         * @return
         */
        public T extractCommitment(DKG.CommitmentMessage commitmentMessage){
            return dkg.decodeCommitment(commitmentMessage.getCommitment().toByteArray());
        }
    }
}
