package meerkat.crypto.dkg.gjkr;

import meerkat.crypto.utils.ChannelImpl;
import meerkat.crypto.utils.Arithmetic;
import meerkat.crypto.utils.concrete.Fp;
import meerkat.crypto.utils.Channel;
import meerkat.crypto.secretsharing.feldman.VerifiableSecretSharing;
import meerkat.crypto.dkg.feldman.DKGMaliciousUser;
import meerkat.crypto.secretsharing.shamir.Polynomial;
import meerkat.crypto.secretsharing.shamir.SecretSharing;
import meerkat.crypto.utils.BigIntegerByteEncoder;
import meerkat.crypto.utils.GenerateRandomPrime;
import org.factcenter.qilin.primitives.Group;
import org.factcenter.qilin.primitives.concrete.Zpstar;
import org.factcenter.qilin.util.ByteEncoder;
import org.junit.Test;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Random;
import java.util.Set;

/**
 * Created by Tzlil on 3/29/2016.
 */
public class SDKGTest {

    int tests = 10;
    BigInteger p = GenerateRandomPrime.SafePrime100Bits;
    BigInteger q = p.subtract(BigInteger.ONE).divide(BigInteger.valueOf(2));
    Group<BigInteger> group = new Zpstar(p);
    Arithmetic<BigInteger> arithmetic = new Fp(q);
    int t = 9;
    int n = 20;


    public void oneTest(Testable testable) throws Exception {
        for (int i = 0; i < testable.threads.length ; i++){
            testable.threads[i].start();
        }
        for (int i = 0; i < testable.threads.length ; i++){
            testable.threads[i].join();
        }

        // got the right public value
        BigInteger publicValue = group.multiply(testable.g,testable.secret);
        for (int i: testable.valids){
            assert (testable.sdkgs[i - 1].getPublicValue().equals(publicValue));
        }

        // assert valid verification values
        BigInteger expected,verification;
        for (int i: testable.valids){
            expected = group.multiply(testable.g, testable.sdkgs[i - 1].getShare().y);
            verification = VerifiableSecretSharing.computeVerificationValue(i, testable.sdkgs[i - 1].getCommitments(), group);
            assert (expected.equals(verification));
        }


        // restore the secret from shares
        ArrayList<Polynomial.Point> sharesList = new ArrayList<Polynomial.Point>();

        for (int i: testable.valids){
            sharesList.add(testable.sdkgs[i - 1].getShare());
        }
        Polynomial.Point[] shares = new Polynomial.Point[sharesList.size()];
        for (int i = 0; i < shares.length; i ++){
            shares[i] = sharesList.get(i);
        }

        BigInteger calculatedSecret = SecretSharing.recoverSecret(shares,arithmetic);
        assert (calculatedSecret.equals(testable.secret));
    }

    public void stopReceivers(Testable testable){
        ChannelImpl channel;
        for (int i = 0 ; i < testable.sdkgs.length ; i++){
            channel = (ChannelImpl)testable.sdkgs[i].getChannel();
            channel.stop();
        }
    }

    @Test
    public void test() throws Exception {
        Testable testable;
        for (int i = 0; i < tests; i++) {
            testable = new Testable(new Random());
            oneTest(testable);
            stopReceivers(testable);
        }
    }
    class Testable{
        Set<Integer> valids;
        Set<Integer> QUAL;
        Set<Integer> aborted;
        Set<Integer> malicious;
        User<BigInteger>[] sdkgs;
        Thread[] threads;
        BigInteger g;
        BigInteger h;
        BigInteger secret;

        public Testable(Random random) {
            this.sdkgs = new User[n];
            this.valids = new HashSet<Integer>();
            this.QUAL = new HashSet<Integer>();
            this.aborted = new HashSet<Integer>();
            this.malicious = new HashSet<Integer>();
            this.threads = new Thread[n];
            this.g = sampleGenerator(random);
            this.h = group.multiply(g,randomIntModQ(random));
            ArrayList<Integer> ids = new ArrayList<Integer>();
            for (int id = 1; id<= n ; id++){
                ids.add(id);
            }
            int id;
            BigInteger s;
            Channel channel;
            Protocol<BigInteger> sdkg;
            this.secret = BigInteger.ZERO;
            ByteEncoder<BigInteger> encoder = new BigIntegerByteEncoder();
            while (!ids.isEmpty()) {
                id = ids.remove(random.nextInt(ids.size()));
                s = randomIntModQ(random);
                channel  = new ChannelImpl(id,n);
                sdkg = new Protocol<BigInteger>(t, n, s, random, q, g , h, group, id,encoder);
                sdkgs[id - 1] = randomSDKGUser(id,channel,sdkg,random);
                threads[id - 1] = new Thread(sdkgs[id - 1]);
                if(QUAL.contains(id)){
                    this.secret =  this.secret.add(s).mod(q);
                }
            }

        }

        public User<BigInteger> randomSDKGUser(int id, Channel channel, Protocol<BigInteger> sdkg, Random random){
            if (QUAL.size() <= t) {
                valids.add(id);
                QUAL.add(id);
                return new User<BigInteger>(sdkg,channel);
            }else{
                int type = random.nextInt(3);
                switch (type){
                    case 0:// regular
                        valids.add(id);
                        QUAL.add(id);
                        return new User<BigInteger>(sdkg,channel);
                    case 1:// abort
                        int abortStage = random.nextInt(3) + 1; // 1 or 2 or 3
                        aborted.add(id);
                        if (abortStage > 1){
                            QUAL.add(id);
                        }
                        return new SDKGUserImplAbort(sdkg,channel,abortStage);
                    case 2:// malicious
                        malicious.add(id);
                        Set<Integer> falls = DKGMaliciousUser.selectFallsRandomly(valids,random);
                        Protocol<BigInteger> maliciousSDKG = SDKGMaliciousUserImpl.generateMaliciousSDKG(sdkg,channel,random);
                        return new SDKGMaliciousUserImpl(sdkg,maliciousSDKG,channel,falls);
                    default:
                        return null;
                }
            }
        }

        public BigInteger sampleGenerator(Random random){
            BigInteger ZERO = group.zero();
            BigInteger g;
            do {
                g = group.sample(random);
            } while (!g.equals(ZERO) && !group.multiply(g, q).equals(ZERO));
            return g;
        }

        public BigInteger randomIntModQ(Random random){
            return new BigInteger(q.bitLength(), random).mod(q);
        }

    }
}
